package nl.wur.bif.pantools.analysis.kmer_classification;

import jakarta.validation.constraints.Max;
import jakarta.validation.constraints.Min;
import nl.wur.bif.pantools.utils.cli.mixins.SelectGenomes;
import nl.wur.bif.pantools.utils.cli.mixins.ThreadNumber;
import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.utils.BeanUtils;

import java.io.IOException;
import java.util.concurrent.Callable;

import static nl.wur.bif.pantools.utils.cli.validation.Constraints.GraphDatabase;
import static nl.wur.bif.pantools.utils.Globals.*;
import static picocli.CommandLine.*;
import static picocli.CommandLine.Model.CommandSpec;

/**
 * Calculate the number of core, accessory, unique, k-mer sequences.
 *
 * @author Robin van Esch, Wageningen University, the Netherlands.
 */
@Command(name = "k_mer_classification", aliases = "kmer_classification", sortOptions = false, abbreviateSynopsis = true)
public class KmerClassificationCLI implements Callable<Integer> {

    @Spec CommandSpec spec;
    @ArgGroup private SelectGenomes selectGenomes;

    @ParentCommand
    @GraphDatabase
    private Pantools pantools;

    @Option(names = {"-p", "--phenotype"})
    private String phenotype;

    @Option(names = "--compressed")
    boolean compressed;

    @Option(names = "--core-threshold")
    @Min(value = 0, message = "{min.ct}")
    @Max(value = 100, message = "{max.ct}")
    int coreThreshold;

    @Option(names = "--unique-threshold")
    @Min(value = 0, message = "{min.ut}")
    @Max(value = 100, message = "{max.ut}")
    int uniqueThreshold;

    @Option(names = "--phenotype-threshold")
    @Min(value = 0, message = "{min.pt}")
    @Max(value = 100, message = "{max.pt}")
    int phenotypeThreshold;

    @Mixin
    private ThreadNumber threadNumber;
    private static int threads;

    @ArgGroup(exclusive = false) SequenceOptions sequenceOptions;
    private static class SequenceOptions {
        @Option(names = {"--sequence"}, required = true)
        void setSequence(boolean value) {sequence = value;}

        @Option(names = "--phasing")
        void setPhasing(boolean value) {phasing = value;}
    }
    private static boolean sequence;
    private static boolean phasing;

    @Override
    public Integer call() throws IOException {
        pantools.createLogger(spec);
        BeanUtils.argValidation(spec, this, selectGenomes);
        threads = threadNumber.getnThreads();
        crossValidation();

        pantools.setPangenomeGraph("pangenome");
        setGlobalParameters(); //TODO: use local parameters instead
        new KmerClassification().kmerClassification(sequence);
        return 0;
    }

    private void setGlobalParameters() {
        setGenomeSelectionOptions(selectGenomes);
        core_threshold = coreThreshold;
        unique_threshold = uniqueThreshold;
        phenotype_threshold = phenotypeThreshold;
        PHENOTYPE = phenotype;
        if (compressed) Mode = "COMPRESSED";
        PHASED_ANALYSIS = phasing;
        THREADS = threads;
    }

    private void crossValidation() {
        if (sequence && threadNumber.getnThreads() < 5 ) {
            Pantools.logger.info("Kmer classification on sequences uses a minium of 5 threads.");
            threads = 5;
        }

        if (!sequence && threadNumber.getnThreads() > 1 ) {
            Pantools.logger.info("Kmer classification on genome level uses only a single thread.");
            threads = 1;
        }
    }
}
