package nl.wur.bif.pantools.analysis.phylogeny.rename_phylogeny;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.analysis.phylogeny.create_tree_template.CreateTreeTemplate;
import nl.wur.bif.pantools.utils.PhasedFunctionalities;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.NotFoundException;
import org.neo4j.graphdb.Transaction;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

import static nl.wur.bif.pantools.analysis.classification.Classification.*;
import static nl.wur.bif.pantools.analysis.phylogeny.Phylogeny.*;
import static nl.wur.bif.pantools.analysis.phylogeny.Phylogeny.renameNexusFile;
import static nl.wur.bif.pantools.utils.PhasedFunctionalities.phasingInfoMap;
import static nl.wur.bif.pantools.utils.Globals.*;
import static nl.wur.bif.pantools.utils.Utils.*;

public class RenamePhylogeny {

    private PhasedFunctionalities pf;

    /**
     * rename_phylogeny()
     *
     * Requires
     * --database-path
     * --input-file a phylogenetic tree
     * --genome or --sequence
     *
     * Optional
     * --phenotype
     * --phased
     * --no-numbers
     *
     * This function only works on the phylogenetic trees generated by pantools. A genome number may only occur once.
     *
     * @param treeType can be "genome", "sequence" or "gene_tree"
     */
    public void renamePhylogeny(boolean excludeGenomeNumbers, String treeType, Path treeFile) {
        Pantools.logger.info("Rename a Newick tree formatted phylogeny.");
        check_database();
        try (Transaction tx = GRAPH_DB.beginTx()) { // start database transaction
            skip.create_skip_arrays(false, true); // create skip array if -skip/-ref is provided by user
            if (PHENOTYPE == null) {
                Pantools.logger.info("No --phenotype provided! Placing only genome numbers at the terminal nodes (leaves).");
            } else {
                Pantools.logger.info("Selected '{}' as phenotype. Using this to rename the terminal nodes (leaves).", PHENOTYPE);
                retrieve_phenotypes(); // Creates and fills geno_pheno_map and phenotype_map when a phenotype was provided by the user
            }
            tx.success(); // transaction successful, commit changes
        } catch (NotFoundException nfe) {
            Pantools.logger.error("Unable to start the database.");
            System.exit(1);
        }

        skip.retrieveSelectedGenomes();
        skip.retrieveSelectedSequences(0, false);
        if (PHASED_ANALYSIS) {
            pf = new PhasedFunctionalities();
            pf.preparePhasedGenomeInformation(true, selectedSequences);
        }

        if (excludeGenomeNumbers) {
            if (PHENOTYPE == null && !treeType.equals("gene_tree")) {
                Pantools.logger.error("--mode no-numbers can only be used in combination with --phenotype.");
                return;
            }
            Pantools.logger.info("Selected --mode no-numbers. Removing the genome numbers from the leaves.");
        }

        if (treeType.equals("sequence")) {
            Pantools.logger.info("--sequence was included, expecting a sequence level tree.");
        } else if (treeType.equals("genome")) {
            Pantools.logger.info("--genome was included, expecting a (sub)genome level tree.");
        } else { // "gene_tree"
            Pantools.logger.info("--gene-tree was included, expecting a gene tree.");
        }

        ArrayList<String> newickOrNexus = readNewickOrNexus(true, treeFile);
        if (newickOrNexus.size() == 1) { // its a newick file
            AtomicInteger changeCounter = new AtomicInteger(0);
            String newick = newickOrNexus.get(0);
            CreateTreeTemplate template = new CreateTreeTemplate();
            ArrayList<String> labels = template.getLabelsFromNewick(treeFile); // collect all identifiers in newick tree
            if (treeType.equals("sequence")) {
                newick = renameNewickOfSequences(labels, excludeGenomeNumbers, newick, changeCounter);
            } else if (treeType.equals("genome")) { // --genome included. tree must have genome numbers or subgenome ids
                newick = renameNewickOfGenomes(labels, excludeGenomeNumbers, newick, changeCounter);
            } else { // gene_tree
                newick = renameNewickOfGenetree(treeFile, excludeGenomeNumbers, newick, changeCounter);
            }

            if (changeCounter.get() == 0) {
                Pantools.logger.info("No changes could be made to the tree.");
            } else {
                String outputFile = addStringBeforeFinalDot(treeFile.toString(), "_RENAMED");
                writeStringToFile(newick, outputFile, false, false);
                Pantools.logger.info("{}/{} leaves were updated", changeCounter.get(), labels.size());
                Pantools.logger.info("New newick written to:");
                Pantools.logger.info(" {}", outputFile);
            }
        } else {
            renameNexusFile(newickOrNexus, excludeGenomeNumbers);
        }
    }

    private String renameNewickOfGenetree(Path treeFile, boolean excludeGenomeNumbers, String phylogeny, AtomicInteger changeCounter) {

        boolean includedGeneCounts = false;
        HashMap<Integer, ArrayList<String>> labelsPerGenome = retrieveTreeLabelsPerGenome(treeFile);
        try (Transaction tx = GRAPH_DB.beginTx()) { // start database transaction
            for (int genomeNr : labelsPerGenome.keySet()) {
                ArrayList<String> updatedLabels = labelsPerGenome.get(genomeNr); // each labels consists of a genome number and the mrna identifier
                String phenotype = create_phenotype_for_rename_phylo(genomeNr);
                String genomeString = getGenomeString(genomeNr, excludeGenomeNumbers);
                if (PHASED_ANALYSIS) {
                    // need to know on which sequence an mRNA is located
                    HashMap<String, Node> mrnaNodeMap = create_mrna_node_map_for_genome(genomeNr, "Replacing identifiers");
                    for (String label : updatedLabels) {
                        String[] labelArray = label.split("_");
                        String mrnaId = getMrnaIdFromlabel(labelArray);
                        Node mrnaNode = mrnaNodeMap.get(mrnaId);
                        if (mrnaNode == null) {
                            System.out.println(label + " cannot be found");
                            continue;
                        }
                        changeCounter.incrementAndGet();
                        int sequenceNr = (int) mrnaNode.getProperty("sequence");
                        String phasingId = createPhasingId(genomeNr + "_" + sequenceNr, true);
                        String[] phylogenyArray = phylogeny.split(genomeString + phasingId + phenotype + mrnaId + ":");
                        if (phylogenyArray.length > 1) {
                            includedGeneCounts = true;
                            phylogeny = phylogeny.replace(label + ":", genomeString + phasingId + phenotype + labelArray[1] + "_" + mrnaId + ":");
                        } else {
                            phylogeny = phylogeny.replace(label + ":", genomeString + phasingId + phenotype + mrnaId + ":");
                        }
                    }
                } else {
                    for (String label : updatedLabels) {
                        changeCounter.incrementAndGet();
                        String[] labelArray = label.split("_");
                        String mrnaId = getMrnaIdFromlabel(labelArray);
                        String[] phylogenyArray = phylogeny.split(genomeString + phenotype + mrnaId + ":");
                        if (phylogenyArray.length > 1) {
                            phylogeny = phylogeny.replace(label + ":", genomeString + phenotype + labelArray[1] + "_" + mrnaId + ":");
                        } else {
                            phylogeny = phylogeny.replace(label + ":", genomeString + phenotype + mrnaId + ":");
                        }
                    }
                }
            }
            tx.success();
        }
        if (includedGeneCounts) {
            Pantools.logger.info("Could not remove all gene counts from the gene tree or else not all identifiers were unique.");
        }
        return phylogeny;
    }

    private String renameNewickOfGenomes(ArrayList<String> labels, boolean excludeGenomeNumbers, String phylogeny,
                                         AtomicInteger changeCounter) {
        for (String label : labels) {
            String[] labelArray = label.split("_");
            int genomeNr = 0;
            try {
                genomeNr = Integer.parseInt(labelArray[0]);
            } catch (NumberFormatException e) {
                Pantools.logger.error("Tree labels should start with a genome number.");
                throw new RuntimeException(e);
            }

            String phenotype = create_phenotype_for_rename_phylo(genomeNr);
            String genomeString = genomeNr +"";
            if (labelArray.length > 1) {
                String[] phases = new String[]{"A","B","C","D","E","F","G","H", "unphased"};
                if (Arrays.asList(phases).contains((labelArray[1])) && excludeGenomeNumbers) {
                    genomeString = labelArray[1]; // still add the phase
                } else if  (Arrays.asList(phases).contains((labelArray[1])) ) {
                    genomeString = "_" + labelArray[1];
                }
            }

            String newLabel;
            if (excludeGenomeNumbers) {
                newLabel = phenotype;
            } else {
                newLabel = genomeString + phenotype;
            }
            phylogeny = renameLabelInNewick(phylogeny, label, newLabel, changeCounter);
        }
        return phylogeny;
    }

    private String renameNewickOfSequences(ArrayList<String> labels, boolean excludeGenomeNumbers, String phylogeny, AtomicInteger changeCounter) {
        for (String label : labels) {
            String[] labelArray = label.split("_");
            int genomeNr = 0;
            try {
                genomeNr = Integer.parseInt(labelArray[0]);
                int sequenceNr = Integer.parseInt(labelArray[1]);
            } catch (NumberFormatException e) {
                Pantools.logger.error("Tree labels should start with a sequence identifier.");
                throw new RuntimeException(e);
            }
            String sequenceId = labelArray[0] + "_" + labelArray[1];
            String phenotype = create_phenotype_for_rename_phylo(genomeNr);
            String phasingId = pf.createPhasingRenamingPhylogeny(sequenceId, true);
            if (excludeGenomeNumbers) {
                sequenceId = "";
            }
            String newLabel = sequenceId + phenotype + phasingId;
            phylogeny = renameLabelInNewick(phylogeny, label, newLabel, changeCounter);
        }
        return phylogeny;
    }

    private String renameLabelInNewick(String phylogeny, String label, String newLabel, AtomicInteger changeCounter) {
        if (phylogeny.contains("(" + label + ":")) {
            phylogeny = phylogeny.replace("(" + label + ":", "(" + newLabel + ":");
            changeCounter.incrementAndGet();
        } else if (phylogeny.contains("(" + label + "_")) {
            String[] phylogeny_array = phylogeny.split("\\(" + label + "_");
            if (phylogeny_array.length != 2) {
                Pantools.logger.error("The provided phylogeny is incorrect. Only trees from one PanTools' phylogenetic methods can be used.");
                System.exit(1);
            }
            int index = phylogeny_array[1].indexOf(':');
            phylogeny = phylogeny_array[0] + "(" + newLabel + phylogeny_array[1].substring(index);
            changeCounter.incrementAndGet();
        } else if (phylogeny.contains("(" + label + ",")) {
            phylogeny = phylogeny.replace("(" + label + ",", "(" + newLabel + ",");
            changeCounter.incrementAndGet();
        } else if (phylogeny.contains("," + label + ")")) {
            phylogeny = phylogeny.replace("," + label + ")", "," + newLabel + ")");
            changeCounter.incrementAndGet();
        } else if (phylogeny.contains("," + label + ":")) {
            phylogeny = phylogeny.replace("," + label + ":", "," + newLabel + ":");
            changeCounter.incrementAndGet();
        } else if (phylogeny.contains("," + label + "_")) {
            String[] phylogeny_array = phylogeny.split("," + label + "_");
            if (phylogeny_array.length != 2) {
                Pantools.logger.error("The provided phylogeny is incorrect. Only trees from one PanTools' phylogenetic methods can be used.");
                System.exit(1);
            }
            int index = phylogeny_array[1].indexOf(':');
            phylogeny = phylogeny_array[0] + "," + newLabel + phylogeny_array[1].substring(index);
            changeCounter.incrementAndGet();
        }
        return phylogeny;
    }

    private String getMrnaIdFromlabel(String[] labelArray) {
        StringBuilder idBuilder = new StringBuilder();
        for (int i = 2; i < labelArray.length; i++) {
            idBuilder.append(labelArray[i]).append("_");
        }
        return idBuilder.toString().replaceFirst(".$",""); // remove last character
    }

    private String getGenomeString(int genomeNr, boolean excludeGenomeNumbers) {
        if (excludeGenomeNumbers) {
            return "";
        }
        return genomeNr + "_";
    }

    /**
     * Return a phenotype value of a specific genome as a string when a --phenotype is included
     * @param sequenceId a sequence identifier: genome number '_' sequence number
     * @return empty when no phased analysis
     */
    private String createPhasingId(String sequenceId, boolean requiresPhasingAnalysisBoolean) {
        String phasingId = "";
        if (PHASED_ANALYSIS || !requiresPhasingAnalysisBoolean) {
            String[] phasingInfo = phasingInfoMap.get(sequenceId);
            if (phasingInfo == null) {
                phasingId = "unphased_";
            } else {
                phasingId = phasingInfo[2].replace("unphased", "_unphased") + "_"; // example [1, B, 1B, 1_B]
            }
        }
        return phasingId;
    }

    private HashMap<Integer, ArrayList<String>> retrieveTreeLabelsPerGenome(Path treeFile) {
        CreateTreeTemplate template = new CreateTreeTemplate();
        ArrayList<String> labels = template.getLabelsFromNewick(treeFile); // collect all identifiers in tree
        HashMap<Integer, ArrayList<String>> labelsPerGenome = new HashMap<>();
        for (String label : labels) {
            String[] labelArray = label.split("_"); // each labels consists of a genome number, gene number (which counts per genome) and the mrna identifier
            int genomeNr = Integer.parseInt(labelArray[0]);
            labelsPerGenome.computeIfAbsent(genomeNr, k -> new ArrayList<>()).add(label);
        }
        return labelsPerGenome;
    }

    /**
     *
     * @return 1 line = newick. multiple lines = nexus
     */
    private ArrayList<String> readNewickOrNexus(boolean print, Path file) {
        ArrayList<String> lines = new ArrayList<>();
        boolean beginTaxafound = false;
        boolean beginSplitsfound = false;
        int lineCounter = 0;

        if (!file.toString().contains(".")) {
            Pantools.logger.error("The provided file must have a file extension, .newick for example.");
            System.exit(1);
        }

        try {
            BufferedReader br = Files.newBufferedReader(file);
            for (String line = br.readLine(); line != null; line = br.readLine()) {
                line = line.trim();
                lines.add(line); // add the regular line before uppercase
                line = line.toUpperCase();
                lineCounter++;
                if (line.startsWith("BEGIN TAXA;")) {
                    beginTaxafound = true;
                } else if (line.startsWith("BEGIN SPLITS;")) {
                    beginSplitsfound = true;
                }
            }
        } catch (IOException ioe) {
            Pantools.logger.error("Failed to read: {}", file);
            System.exit(1);
        }

        if (lineCounter == 1) { // assume file is newick format, no file checks
            if (print) {
                Pantools.logger.info("Input file was recognized as newick.");
            }
        } else if (beginTaxafound && beginSplitsfound) {
            if (print) {
                Pantools.logger.info("Input file was recognized as nexus.");
            }
        } else {
            Pantools.logger.error("File could not be recognized as newick or nexus.");
            System.exit(1);
        }
        return lines;
    }

    /**
     * Return a phenotype value of a specific genome as a string when a --phenotype is included
     * @param genome_nr
     * @return
     */
    private String create_phenotype_for_rename_phylo(int genome_nr) {
        String phenotype = "";
        if (PHENOTYPE == null) {
            return phenotype;
        }
        if (geno_pheno_map.containsKey(genome_nr)) {
            phenotype = geno_pheno_map.get(genome_nr);
        }
        if (phenotype.equals("?")) {
            phenotype = "Unknown";
        }
        if (!phenotype.equals("") && Mode.equals("0")) {
            phenotype = "_" + phenotype;
        }
        return phenotype;
    }
}
