package nl.wur.bif.pantools.construction.index;

import nl.wur.bif.pantools.Pantools;
import nl.wur.bif.pantools.construction.sequence.SequenceDatabase;
import org.neo4j.graphdb.GraphDatabaseService;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.ResourceIterator;
import org.neo4j.graphdb.Transaction;

import java.io.*;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.logging.Level;
import java.util.logging.Logger;

import static nl.wur.bif.pantools.utils.Globals.*;
import static nl.wur.bif.pantools.utils.Utils.runCommandWithOutput;

/**
 * Implements all the functionality to build a KMC-based kmer index. 
 * index database. 
 * 
 * @author Siavash Sheikhizadeh, Bioinformatics chairgroup, Wageningen
 * University, the Netherlands
 */
public final class IndexDatabase {

    public int MAX_BYTE_COUNT = 1 << 24;// 16 MB 
    public int header_pos;
    public final String INFORMATION_FILE_EXTENTION = ".info";
    public final String PREFIX_FILE_EXTENTION = ".kmc_pre";
    public final String SUFFIX_FILE_EXTENTION = ".kmc_suf";
    public final String POINTER_FILE_EXTENTION = ".pointer";

    public static int K;
    public int mode;
    public int ctr_size;
    public int pre_len;
    public int min_count;
    public int max_count;
    public long kmers_num;
    public int id_len;
    public int offset_len;

    public int suf_len;
    public  int poniter_length;
    public long[] prefix_ptr;
    public RandomAccessFile suf_file;
    public RandomAccessFile pre_file;
    public MappedByteBuffer[] suf_buff;
    public RandomAccessFile ptr_file;
    public MappedByteBuffer[] ptr_buff;
    public MappedByteBuffer pre_buff[];
    public String KMC = "kmc";

    /**
     * Mounts an available index database to the index database object
     * 
     * @param index_path Path to the index database 
     * @param index_name The prefix of the KMC database
     */
    public IndexDatabase(String index_path, String index_name) {
        int i, j, number_of_pages, record_size, page_size, full_page_size;
        long number_of_prefixes;
        try {
            //Pantools.logger.info("Mounting index {}", index_path);
            pre_file = new RandomAccessFile(index_path + index_name + PREFIX_FILE_EXTENTION, "r");
            BufferedReader in = new BufferedReader(new FileReader(index_path + index_name + INFORMATION_FILE_EXTENTION));
            K = Integer.parseInt(in.readLine().trim().split(":")[1]);
            mode = Integer.parseInt(in.readLine().trim().split(":")[1]);
            ctr_size = Integer.parseInt(in.readLine().trim().split(":")[1]);
            pre_len = Integer.parseInt(in.readLine().trim().split(":")[1]);
            min_count = Integer.parseInt(in.readLine().trim().split(":")[1]);
            max_count = Integer.parseInt(in.readLine().trim().split(":")[1]);
            kmers_num = Long.valueOf(in.readLine().trim().split(":")[1]);
            suf_len = Integer.parseInt(in.readLine().trim().split(":")[1]);
            id_len = Integer.parseInt(in.readLine().trim().split(":")[1]);
            offset_len = Integer.parseInt(in.readLine().trim().split(":")[1]);
            poniter_length = Integer.parseInt(in.readLine().trim().split(":")[1]);
            in.close();
        // load the prefix file into the memory
            Pantools.logger.trace("Loading existing prefixes in memory...");
            number_of_prefixes = 1 << (2 * pre_len);
            prefix_ptr = new long[(int)number_of_prefixes];
            record_size = 8; // in bytes
            full_page_size = MAX_BYTE_COUNT / record_size * record_size; // in bytes
            number_of_pages = (int) ((number_of_prefixes * record_size) / full_page_size + ((number_of_prefixes * record_size) % full_page_size == 0 ? 0 : 1));
            pre_buff = new  MappedByteBuffer[number_of_pages];
            Pantools.logger.trace("pre_len {}", pre_len);
            Pantools.logger.trace("number_of_prefixes {}", number_of_prefixes);
            Pantools.logger.trace("full_page_size {}", full_page_size);
            Pantools.logger.trace("number_of_pages {}", number_of_pages);
            for (i = 0; i < number_of_pages; ++i) {
                page_size = (int) (i == number_of_pages - 1 ? (number_of_prefixes * record_size) % full_page_size : full_page_size); // in bytes
                page_size = page_size == 0 ? full_page_size : page_size; // in bytes
                Pantools.logger.trace("page_size  {}", page_size);
                pre_buff[i] = pre_file.getChannel().map(FileChannel.MapMode.READ_ONLY, ((long)full_page_size) * i + 4, page_size);
                pre_buff[i].position(0);
                for (j = 0; j < page_size / 8; ++j) 
                    prefix_ptr[(int)(((long)full_page_size) * i / 8 + j)] = read_prefix(pre_buff[i]);
                pre_buff[i] = null;
            }
            pre_buff = null;
            pre_file.close();
        // mapping suffix file into the memory
            Pantools.logger.trace("Mapping existing suffix file in memory...");
            record_size = ctr_size + suf_len / 4; // in bytes
            full_page_size = MAX_BYTE_COUNT / record_size * record_size; // in bytes
            number_of_pages = (int) ((kmers_num * record_size) / full_page_size + ((kmers_num * record_size) % full_page_size == 0 ? 0 : 1));
            suf_file = new RandomAccessFile(index_path + index_name + SUFFIX_FILE_EXTENTION, "r");
            suf_buff = new MappedByteBuffer[number_of_pages];
            for (i = 0; i < number_of_pages; ++i) {
                page_size = (int) (i == number_of_pages - 1 ? (kmers_num * record_size) % full_page_size : full_page_size); // in bytes
                page_size = page_size == 0 ? full_page_size : page_size; // in bytes
                suf_buff[i] = suf_file.getChannel().map(FileChannel.MapMode.READ_ONLY, ((long)full_page_size) * i + 4, page_size);
            }
        // mapping pointers file into the memory
            Pantools.logger.trace("Mapping existing pointers file in memory...");
            record_size = poniter_length; // in bytes
            full_page_size = MAX_BYTE_COUNT / record_size * record_size; // in bytes
            number_of_pages = (int) ((kmers_num * record_size) / full_page_size + ((kmers_num * record_size) % full_page_size == 0 ? 0 : 1));
            ptr_file = new RandomAccessFile(index_path + index_name + POINTER_FILE_EXTENTION, "rw");
            ptr_buff = new MappedByteBuffer[number_of_pages];
            for (i = 0; i < number_of_pages; ++i) {
                page_size = (int) (i == number_of_pages - 1 ? (kmers_num * record_size) % full_page_size : full_page_size); // in bytes
                page_size = page_size == 0 ? full_page_size : page_size; // in bytes
                ptr_buff[i] = ptr_file.getChannel().map(FileChannel.MapMode.READ_WRITE, ((long)full_page_size) * i + 4, page_size);
            }
        } catch (IOException e) {
            Pantools.logger.error(e.getMessage());
            System.exit(1);
        }
    }

    /**
     *  Creates a new index database for a set of genomes.
     * 
        All integers in the KMC output files are stored in LSB (least significant byte first) format.
        A .kmc pre file contains, in the order:
        [marker] : 4 bytes with the letters: KMCP.
        [data]: Here are k-mer prefix data. More precisely, it is an array of uint64 elements, of size 4^prefix length. Position x
                in the array stores the number of different k-mers (in binary) whose prefix is less
                than x. DNA symbols are encoded as follows: A   0, C   1, G   2, T   3. For example, if the queried
                k-mer is ACGACTGA and lut prefix length = 4, then we cut off the first 4 symbols, i.e., the prefix ACGA,
                which is interpreted in binary as x = 24 (since 0 (* 2^6 + 1 * 2^4 + 2 * 2^2 + 0 * 2^0 = 24). Now we look into
                “data” at locations x and x+1 to read, e.g., 1523 and 1685. This means that in the file .kmc suf in records from
                1523 to 1685􀀀1 there are suffixes of the k-mers with prefix ACGA. Having got this range, we can now binary
                search the suffix CTGA.            
        [header] : 
        [header position] : An integer consisting of the last 4 bytes in the file. Its contains the relative position of the beginning of the
        field [header].
        [marker] : another copy, to signal the file is not truncated.
        
        After opening the file, one should do the following:
        1. Read the first 4 bytes and check if they contain the letters KMCP.
        2. Jump to position 4 bytes back from end of file (excluding the marker) and read the header position x.
        3. Jump to position x + 4 (excluding the marker) bytes back from end of file and read the header.
        4. Read [data].
            
        A .kmc suf file contains, in order:
        [marker] : 4 bytes with the letters: KMCS
        [data] :An array record t records[total kmers].
                total kmers is a value taken from the .kmc pre file.
                record t is a type representing a k-mer. Its first field is the k-mer suffix string, stored on (kmer length 􀀀
                lut prefix length)=4 bytes. The next field is counter size, with the number of bytes used by the counter,
                which is either a 1 : : : 4-byte integer, or a 4-byte float.            
                The k-mers are stored with their leftmost symbol first, packed into bytes. For example, CCACAAAT is
                represented as 0x51 (for CCAC), 0x03 (for AAAT). Integers are stored according to the LSB (little endian)
                convention, floats are stored in the same way as they are stored in the memory.
        [marker] (another copy, to signal the file is not truncated).
     * 
     * @param index_path Path to the index database 
     * @param genomes_path_file A text file containing path to the genomes
     * @param genomeDb The genome database
     * @param k Size of K
     */
    public IndexDatabase(String index_path, String genomes_path_file, SequenceDatabase genomeDb, int k) {
        int i, j, number_of_pages, record_size, page_size, full_page_size;
        long number_of_prefixes, longest_scaffold = 0;
        Pantools.logger.debug("Creating index in {}", index_path);
        try {
            Files.createDirectory(Paths.get(index_path));
            Run_KMC(index_path, genomes_path_file, genomeDb, k);
            pre_file = new RandomAccessFile(index_path + "sorted" + PREFIX_FILE_EXTENTION, "r");
            pre_file.seek(pre_file.length() - 8);
            header_pos = read_int(pre_file);
            pre_file.seek(pre_file.length() - (8 + header_pos));
        // read the header of the index    
            K = read_int(pre_file);
        //  k-mer length   
            mode = read_int(pre_file); 
        // 0 (occurrence count) or 1 (counting according to Quake quality)
            ctr_size = read_int(pre_file); 
        // counter field size: for mode 0 it is 1, 2, 3, or 4; for mode 1 it is always 4
            pre_len = read_int(pre_file);  
        // prefix length such that suffix length is divisible by 4. 
        // Max prefix length is limited up to 15 in KMC (exact value is calculated in such a way that summary size of kmc_pre and kmc_suf should be minimal)
            min_count = read_int(pre_file);
        // minimum number of k-mer occurrences to be written in the database    
            max_count = read_int(pre_file);
        // maximum number of k-mer occurrences to be written in the database    
            kmers_num = read_long(pre_file);
        // total number of k-mers in the database    
            suf_len = K - pre_len;
            for (i = 1; i <= genomeDb.num_genomes; ++i)
                for (j = 1; j <= genomeDb.num_sequences[i]; ++j)
                    if (genomeDb.sequence_length[i][j] > longest_scaffold)
                        longest_scaffold = genomeDb.sequence_length[i][j];
            id_len = (int)Math.round(Math.ceil( Math.log(kmers_num) / Math.log(2) / 8));
            offset_len = (int)Math.round(Math.ceil( Math.log(longest_scaffold) / Math.log(2) / 8));
            poniter_length = 2 * id_len + offset_len + 1;
            write_info(index_path, "sorted");
            Pantools.logger.debug("Indexing kmers...");
        // load the prefix file into the memory    
            Pantools.logger.trace("Loading prefixes in memory...");
            number_of_prefixes = 1 << (2 * pre_len);
            prefix_ptr = new long[(int)number_of_prefixes];
            record_size = 8; // in bytes
            full_page_size = MAX_BYTE_COUNT / record_size * record_size; // in bytes
            number_of_pages = (int) ((number_of_prefixes * record_size) / full_page_size + ((number_of_prefixes * record_size) % full_page_size == 0 ? 0 : 1));
            pre_buff = new  MappedByteBuffer[number_of_pages];
            for (i = 0; i < number_of_pages; ++i) {
                page_size = (int) (i == number_of_pages - 1 ? (number_of_prefixes * record_size) % full_page_size : full_page_size); // in bytes
                page_size = page_size == 0 ? full_page_size : page_size; // in bytes
                Pantools.logger.trace("page_size  {}", page_size);
                pre_buff[i] = pre_file.getChannel().map(FileChannel.MapMode.READ_ONLY, ((long)full_page_size) * i + 4, page_size);
                for (j = 0; j < page_size / 8; ++j) 
                    prefix_ptr[(int)(((long)full_page_size) * i / 8 + j)] = read_prefix(pre_buff[i]);
                pre_buff[i] = null;
            }
            pre_buff = null;
            pre_file.close();
        // mapping suffix file into the memory
            Pantools.logger.trace("Mapping suffixes file to memory...");
            record_size = ctr_size + suf_len / 4; // in bytes
            full_page_size = MAX_BYTE_COUNT / record_size * record_size; // in bytes
            number_of_pages = (int) ((kmers_num * record_size) / full_page_size + ((kmers_num * record_size) % full_page_size == 0 ? 0 : 1));
            suf_file = new RandomAccessFile(index_path + "sorted" + SUFFIX_FILE_EXTENTION, "r");
            suf_buff = new MappedByteBuffer[number_of_pages];
            for (i = 0; i < number_of_pages; ++i) {
                page_size = (int) (i == number_of_pages - 1 ? (kmers_num * record_size) % full_page_size : full_page_size); // in bytes
                page_size = page_size == 0 ? full_page_size : page_size; // in bytes
                suf_buff[i] = suf_file.getChannel().map(FileChannel.MapMode.READ_ONLY, ((long)full_page_size) * i + 4, page_size);
            }
        // mapping pointers file into the memory
            Pantools.logger.trace("Mapping pointers file to memory...");
            record_size = poniter_length; // in bytes
            full_page_size = MAX_BYTE_COUNT / record_size * record_size; // in bytes
            number_of_pages = (int) ((kmers_num * record_size) / full_page_size + ((kmers_num * record_size) % full_page_size == 0 ? 0 : 1));
            ptr_file = new RandomAccessFile(index_path + "sorted" + POINTER_FILE_EXTENTION, "rw");
            ptr_buff = new MappedByteBuffer[number_of_pages];
            for (i = 0; i < number_of_pages; ++i) {
                page_size = (int) (i == number_of_pages - 1 ? (kmers_num * record_size) % full_page_size : full_page_size); // in bytes
                page_size = page_size == 0 ? full_page_size : page_size; // in bytes
                ptr_buff[i] = ptr_file.getChannel().map(FileChannel.MapMode.READ_WRITE, ((long)full_page_size) * i + 4, page_size);
                for (j = 0; j < page_size; ++j)
                    ptr_buff[i].put((byte)-1);
            }
        } catch (IOException e) {
            Pantools.logger.error(e.getMessage());
            System.exit(1);
        }
    }
       
    /**
     * Updates an available index database to include new genomes
     * 
     * @param index_path Path to the index database 
     * @param genomes_path_file A text file containing path to the new genomes
     * @param genomeDb The genome database
     * @param graphDb The graph database
     * @param old_database The existing index database
     * @param previous_num_genomes Number of the genomes available in the index
     */
    public IndexDatabase(String index_path, String genomes_path_file, SequenceDatabase genomeDb, GraphDatabaseService graphDb, IndexDatabase old_database, int previous_num_genomes) {
        int i, j, number_of_pages, record_size, page_size, full_page_size;
        long number_of_prefixes, longest_scaffold = 0;
        long c_index, p_index, l;
        Node node;
        ResourceIterator<Node> nodes_iterator;
        // move current index files to directory old_darabase
       
        try{
            // make new index for new genomes
            Pantools.logger.info("Running KMC with K = {} ...", old_database.K);
            final String index_command = KMC + " -cs127 -r -k" + old_database.K + " -t" + THREADS + " -ci1 -fm " + (genomeDb.num_genomes - previous_num_genomes > 1 ? "@" +
                    genomes_path_file.trim() : genomeDb.genome_names[previous_num_genomes + 1]) +
                    " " + index_path + "new " + index_path;
            Pantools.logger.debug("KMC command: {}", index_command);

            final String index_output = runCommandWithOutput(index_command);
            Pantools.logger.trace("KMC output: {}", index_output);

            // merge two indices
            Pantools.logger.debug("Merging two indices.");

            final String merge_command = KMC + "_tools simple " + index_path + "sorted " + index_path + "new union " + index_path + "merged";
            Pantools.logger.debug("KMC command: {}", merge_command);

            final String merge_output = runCommandWithOutput(merge_command);
            Pantools.logger.trace("KMC output: {}", merge_output);

            pre_file = new RandomAccessFile(index_path + "merged" + PREFIX_FILE_EXTENTION, "r");
            pre_file.seek(pre_file.length() - 8);
            header_pos = read_int(pre_file);
            pre_file.seek(pre_file.length() - 8 - header_pos);

            // read the merged index properties
            K = read_int(pre_file);
            mode = read_int(pre_file);
            ctr_size = read_int(pre_file);
            pre_len = read_int(pre_file);
            min_count = read_int(pre_file);
            max_count = read_int(pre_file);
            kmers_num = read_long(pre_file);
            suf_len = K - pre_len;
            for (i = 1; i <= genomeDb.num_genomes; ++i)
                for (j = 1; j <= genomeDb.num_sequences[i]; ++j)
                    if (genomeDb.sequence_length[i][j] > longest_scaffold)
                        longest_scaffold = genomeDb.sequence_length[i][j];
            id_len = (int)Math.round(Math.ceil( Math.log(kmers_num) / Math.log(2) / 8));
            offset_len = (int)Math.round(Math.ceil( Math.log(longest_scaffold) / Math.log(2) / 8));
            poniter_length = 2 * id_len + offset_len + 1;
            write_info(index_path, "merged");
            Pantools.logger.info("Number of available kmers:\t{}", old_database.kmers_num);
            Pantools.logger.info("Number of new kmers:\t{}", (kmers_num - old_database.kmers_num));
            Pantools.logger.info("Total number of kmers:\t{}", kmers_num);

            // load the prefix file into the memory
            Pantools.logger.trace("loading new prefixes in memory...");
            number_of_prefixes = 1 << (2 * pre_len);
            prefix_ptr = new long[(int)number_of_prefixes];
            record_size = 8; // in bytes
            full_page_size = MAX_BYTE_COUNT / record_size * record_size; // in bytes
            number_of_pages = (int) ((number_of_prefixes * record_size) / full_page_size + ((number_of_prefixes * record_size) % full_page_size == 0 ? 0 : 1));
            pre_buff = new  MappedByteBuffer[number_of_pages];
            for (i = 0; i < number_of_pages; ++i) {
                page_size = (int) (i == number_of_pages - 1 ? (number_of_prefixes * record_size) % full_page_size : full_page_size); // in bytes
                page_size = page_size == 0 ? full_page_size : page_size; // in bytes
                pre_buff[i] = pre_file.getChannel().map(FileChannel.MapMode.READ_ONLY, ((long)full_page_size) * i + 4, page_size);
                for (j = 0; j < page_size / 8; ++j) 
                    prefix_ptr[(int)(((long)full_page_size) * i / 8 + j)] = read_prefix(pre_buff[i]);
                pre_buff[i] = null;
            }
            pre_buff = null;
            pre_file.close();

            // mapping suffix file into the memory
            Pantools.logger.trace("Mapping new suffix file to memory...");
            record_size = ctr_size + suf_len / 4;
            full_page_size = MAX_BYTE_COUNT / record_size * record_size; // in bytes
            number_of_pages = (int) ((kmers_num * record_size) / full_page_size + ((kmers_num * record_size) % full_page_size == 0 ? 0 : 1));
            suf_file = new RandomAccessFile(index_path + "merged" + SUFFIX_FILE_EXTENTION, "r");
            suf_buff = new MappedByteBuffer[number_of_pages];
            for (i = 0; i < number_of_pages; ++i) {
                page_size = (int) (i == number_of_pages - 1 ? (kmers_num * record_size) % full_page_size : full_page_size); // in bytes
                page_size = page_size == 0 ? full_page_size : page_size; // in bytes
                suf_buff[i] = suf_file.getChannel().map(FileChannel.MapMode.READ_ONLY, ((long)full_page_size) * i + 4, page_size);
            }
        // mapping pointers file into the memory
            Pantools.logger.trace("Mapping new pointer files to memory...");
            record_size = poniter_length; // in bytes
            full_page_size = MAX_BYTE_COUNT / record_size * record_size; // in bytes
            number_of_pages = (int) ((kmers_num * record_size) / full_page_size + ((kmers_num * record_size) % full_page_size == 0 ? 0 : 1));
            ptr_file = new RandomAccessFile(index_path + "merged" + POINTER_FILE_EXTENTION, "rw");
            ptr_buff = new MappedByteBuffer[number_of_pages];
            for (i = 0; i < number_of_pages; ++i) {
                page_size = (int) (i == number_of_pages - 1 ? (kmers_num * record_size) % full_page_size : full_page_size); // in bytes
                page_size = page_size == 0 ? full_page_size : page_size; // in bytes
                ptr_buff[i] = ptr_file.getChannel().map(FileChannel.MapMode.READ_WRITE, ((long)full_page_size) * i + 4, page_size);
                for (j = 0; j < page_size; ++j)
                    ptr_buff[i].put((byte)-1);
            }
        // adjusting available pointers
            Pantools.logger.info("Updating kmer index...");
            kmer old_kmer = new kmer(K, old_database.pre_len);
            kmer new_kmer = new kmer(K, pre_len);
            IndexScanner old_scanner = new IndexScanner(old_database);
            IndexScanner new_scanner = new IndexScanner(this);
            IndexPointer ptr = new IndexPointer();
            try(Transaction tx = graphDb.beginTx()) {
                nodes_iterator = graphDb.findNodes(NUCLEOTIDE_LABEL);
                for(;nodes_iterator.hasNext();) {
                    node = nodes_iterator.next();
                    if (!node.hasLabel(DEGENERATE_LABEL)) {
                        l = (long)node.getProperty("first_kmer");
                        old_scanner.get_kmer(old_kmer, l);
                        kmer.copy_and_adjust(new_kmer, old_kmer);
                        p_index = new_scanner.find(new_kmer);
                        old_scanner.get_pointer(ptr,l);
                        new_scanner.put_pointer(ptr,p_index);
                        node.setProperty("first_kmer", p_index);
                        for(l = old_scanner.get_next_index(l); l != -1L; l = old_scanner.get_next_index(l)) {
                            old_scanner.get_kmer(old_kmer, l);
                            kmer.copy_and_adjust(new_kmer, old_kmer);
                            c_index = new_scanner.find(new_kmer);
                            old_scanner.get_pointer(ptr,l);
                            new_scanner.put_pointer(ptr,c_index);
                            new_scanner.put_next_index(c_index,p_index);
                            p_index = c_index;
                        }
                        new_scanner.put_next_index(-1L,p_index);
                        node.setProperty("last_kmer", p_index);
                    }
                }
                nodes_iterator.close();
                tx.success();
            }
            old_database.close();
            Files.delete(Paths.get(index_path + "sorted" + PREFIX_FILE_EXTENTION));
            Files.delete(Paths.get(index_path + "sorted" + SUFFIX_FILE_EXTENTION));
            Files.delete(Paths.get(index_path + "sorted" + INFORMATION_FILE_EXTENTION));
            Files.delete(Paths.get(index_path + "sorted" + POINTER_FILE_EXTENTION));
            Files.delete(Paths.get(index_path + "new" + PREFIX_FILE_EXTENTION));
            Files.delete(Paths.get(index_path + "new" + SUFFIX_FILE_EXTENTION));
            Files.move(Paths.get(index_path + "merged" + PREFIX_FILE_EXTENTION), Paths.get(index_path + "sorted" + PREFIX_FILE_EXTENTION));
            Files.move(Paths.get(index_path + "merged" + SUFFIX_FILE_EXTENTION), Paths.get(index_path + "sorted" + SUFFIX_FILE_EXTENTION));
            Files.move(Paths.get(index_path + "merged" + POINTER_FILE_EXTENTION), Paths.get(index_path + "sorted" + POINTER_FILE_EXTENTION));
            Files.move(Paths.get(index_path + "merged" + INFORMATION_FILE_EXTENTION), Paths.get(index_path + "sorted" + INFORMATION_FILE_EXTENTION));
        } catch (IOException e) {
            Pantools.logger.error("Failed to merge index databases.");
            throw new RuntimeException("Failed to make index", e);
        }
    }

    /**
     * Runs KMC with the given or optimum K value and creates KMC database. 
     * @param index_path Path to the index database
     * @param genomes_path_file A text file containing path to the genomes
     * @param genomeDb The graph database
     * @param k  Size of K (is -1 if K is not given by the user)
     */
    public void Run_KMC(String index_path, String genomes_path_file, SequenceDatabase genomeDb, int k) {
        double p = 0;
        try{
        Runtime.getRuntime().exec(KMC); // to check if kmc is reachable
        if (k > -1) { // K is not given by the user, then calculate the optimal K
            Pantools.logger.info("K is given by the user: {}", k);

            K = k;
            if (K % 2 == 0) // Even values are problamatic to localization process 
                K += 1;

            final String kmc_command = KMC + " -cs127 -k" + K + " -t" + THREADS + " -ci1 -fm " +
                    (genomeDb.num_genomes > 1 ? "@" + genomes_path_file.trim() : genomeDb.genome_names[1]) + " " + index_path + "/kmers " + index_path;
            Pantools.logger.debug("KMC command: {}", kmc_command);
            final String output = runCommandWithOutput(kmc_command);
            Pantools.logger.trace("KMC output: {}", output);

            if (checkIfKmcIndexExists(index_path)) {
                sortKmcIndex(index_path);
            } else {
                Pantools.logger.error("No kmc index found in {}", index_path);
                throw new RuntimeException("Missing kmc index");
            }
        } else {
            Pantools.logger.info("K is not given by the user. Calculating the optimal K.");

            K = Math.round((float)(Math.log(genomeDb.number_of_bytes * 200)/Math.log(4)));
            K += (K % 2 == 0 ? 1 : 0);
            do {
                K -= 2;
                Pantools.logger.debug("K = {}", K);

                final String kmc_command = KMC + " -cs127 -k" + K + " -t" + THREADS + " -ci1 -fm " +
                        (genomeDb.num_genomes > 1 ? "@" + genomes_path_file.trim() : genomeDb.genome_names[1]) + " " + index_path + "/kmers " + index_path;
                Pantools.logger.debug("KMC command: {}", kmc_command);

                final String kmc_output = runCommandWithOutput(kmc_command);
                Pantools.logger.trace("KMC output: {}", kmc_output);

                if (checkIfKmcIndexExists(index_path)) {
                    sortKmcIndex(index_path);

                    pre_file = new RandomAccessFile(index_path + "/sorted.kmc_pre", "r");
                    pre_file.seek(pre_file.length() - 8);
                    header_pos = read_int(pre_file);
                    pre_file.seek(pre_file.length() - (8 + header_pos));
                    // read the header of the index    
                    K = read_int(pre_file);
                    //  k-mer length
                    mode = read_int(pre_file);
                    // 0 (occurrence count) or 1 (counting according to Quake quality)
                    ctr_size = read_int(pre_file);
                    // counter field size: for mode 0 it is 1, 2, 3, or 4; for mode 1 it is always 4
                    pre_len = read_int(pre_file);
                    // prefix length such that suffix length is divisible by 4.
                    // Max prefix length is limited up to 15 in KMC (exact value is calculated in such a way that summary size of kmc_pre and kmc_suf should be minimal)
                    min_count = read_int(pre_file);
                    // minimum number of k-mer occurrences to be written in the database    
                    max_count = read_int(pre_file);
                    // maximum number of k-mer occurrences to be written in the database
                    kmers_num = read_long(pre_file);  
                    p = (double)kmers_num / (1l  << (2 * K)) * 2;// 2n / 4 ^ K
                    pre_file.close();
                } else {
                    Pantools.logger.error("No kmc index found in {}", index_path);
                    throw new RuntimeException("Missing kmc index");
                }
            } while (p < 0.01);
            K += 2;

            final String kmc_command = KMC + " -cs127 -k" + K + " -t" + THREADS + " -ci1 -fm " +
                    (genomeDb.num_genomes > 1 ? "@" + genomes_path_file.trim() : genomeDb.genome_names[1]) + " " + index_path + "/kmers " + index_path;
            Pantools.logger.debug("KMC command: {}", kmc_command);

            final String kmc_output = runCommandWithOutput(kmc_command);
            Pantools.logger.trace("KMC output: {}", kmc_output);

            if (checkIfKmcIndexExists(index_path)) {
                sortKmcIndex(index_path);
            } else {
                Pantools.logger.error("No kmc index found in {}", index_path);
                throw new RuntimeException("Missing kmc index");
            }
        }
        Pantools.logger.debug("K = {}", K);
        } catch (IOException ex) {
            Pantools.logger.warn("Failed to find K!");
        }
    }

    /**
     * Checks if the KMC index exists in the given path
     * @param index_path Path to the index database
     * @return True if the KMC index exists, false otherwise
     */
    private boolean checkIfKmcIndexExists(String index_path) {
        Pantools.logger.debug("Checking if kmc index exists in {}", index_path);
        boolean indexExists = new File(index_path + "/kmers.kmc_pre").exists() && new File(index_path + "/kmers.kmc_suf").exists();
        Pantools.logger.trace("KMC index exists: {}", indexExists);
        return indexExists;
    }

    /**
     * Sorts the KMC index in the given path
     * @param index_path Path to the index database
     */
    private void sortKmcIndex(String index_path) {
        Pantools.logger.debug("Sorting kmc index in {}", index_path);

        final String command = KMC + "_tools transform " + index_path + "/kmers sort " + index_path + "/sorted";
        Pantools.logger.debug("KMC command: {}", command);

        final String output = runCommandWithOutput(command);
        Pantools.logger.trace("KMC output: {}", output);

        // Small databases are usually sorted already
        if (output.startsWith("Warning: input database is already sorted.")) {
            Pantools.logger.trace("KMC index is already sorted.");
            new File(index_path + "/kmers.kmc_pre").renameTo(new File(index_path + "/sorted.kmc_pre"));
            new File(index_path + "/kmers.kmc_suf").renameTo(new File(index_path + "/sorted.kmc_suf"));
        } else {
            new File(index_path + "/kmers.kmc_pre").delete();
            new File(index_path + "/kmers.kmc_suf").delete();
        }
    }

    /**
     * Writes the index database specification in the info file of the index database
     * 
     * @param index_path Path to the index database
     * @param index_name The prefix of the KMC database
     */
    public void write_info(String index_path, String index_name) {
        try {
            BufferedWriter out = new BufferedWriter(new FileWriter(index_path + index_name + INFORMATION_FILE_EXTENTION));
            out.write("K:" + K + "\n");
            out.write("mode:" + mode + "\n");
            out.write("ctr_size:" + ctr_size + "\n");
            out.write("pre_len:" + pre_len + "\n");
            out.write("min_count:" + min_count + "\n");
            out.write("max_count:" + max_count + "\n");
            out.write("kmers_num:" + kmers_num + "\n");
            out.write("suf_len:" + suf_len + "\n");
            out.write("id_len:" + id_len + "\n");
            out.write("offset_len:" + offset_len + "\n");
            out.write("poniter_length:" + poniter_length + "\n");
            out.close();
        } catch (IOException ioe) {
            Pantools.logger.error(ioe.getMessage());
            System.exit(1);
        }
    }

    /**
     * @return Number of kmers in the KMC database
     */
    public long length() {
        return kmers_num;
    }

    /**
     * @return Prefix length of kmers
     */
    public int get_pre_len() {
        return pre_len;
    }

    /**
     * @return suffix length of kmers
     */
    public int get_suf_len() {
        return suf_len;
    }

    /**
     * Closes the index database.
     */
    public void close() {
        try {
            ptr_file.close();
            suf_file.close();
            int k;
            for (k = 0; k < ptr_buff.length; ++k) {
                ptr_buff[k] = null;
            }
            for (k = 0; k < suf_buff.length; ++k) {
                suf_buff[k] = null;
            }
        } catch (IOException e) {
            Pantools.logger.error(e.getMessage());
            System.exit(1);
        }
    }
    
    /**
     * Reads a long integer from the current position of the memory mapped buffer.
     * 
     * @param buff The memory mapped buffer
     * @return Long integer read from the current position of the memory mapped buffer.
     */
    private long read_prefix(MappedByteBuffer buff) {
        int i;
        long number = 0;
        for (i = 0; i < 8; ++i)
            number += ((long)(buff.get() & 0x00FF)) << (8 * i);
        return number;
    }    

    /**
     * Reads an integer from current position of the byte stream.
     * 
     * @param file The file associated to the byte stream
     * @return Integer read from the current position of the stream.
     */
    private int read_int(RandomAccessFile file) {
        int number = 0;
        for (int i = 0; i < 4; ++i) {
            try {
                number += ((int)(file.readByte() & 0x00FF) << (8 * i));
            } catch (IOException ex) {
                Logger.getLogger(IndexDatabase.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        return number;
    }

    /**
     * Reads a long integer from current position of the byte stream.
     * 
     * @param file The file associated to the byte stream
     * @return Long integer read from the current position of the stream.
     */
    private long read_long(RandomAccessFile file) throws IOException {
        long number = 0;
        for (int i = 0; i < 8; ++i) {
            number += ((long)(file.readByte() & 0x00FF) << (8 * i));
        }
        return number;
    }

    /**
     * @return The size of K that KMC database has been built upon
     */
    public int get_K() {
        return K;
    }
    
    /**
     * @return The length of records in the pointers database in bytes
     */
    public int get_pointer_length() {
        return poniter_length;
    }
}
