"""
Compare output files generated by PanTools with cmp.
Will exit with a non-zero exit code if files are different.
"""

rule compare_exported_csvs:
    """Compare sorted CSV files from export_pangenome."""
    input:
        "output/exports/{dataset_name}_{dataset_type}/{version1}/{comparison_type}.sorted.csv",
        "output/exports/{dataset_name}_{dataset_type}/{version2}/{comparison_type}.sorted.csv"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/exports/{comparison_type}.csv.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_proteins:
    """Compare proteome files created by pantools add_annotations."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/proteins/{filename}.fasta",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/proteins/{filename}.fasta"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/proteins/{filename}.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_sams:
    """
    Compare SAM files from read mapping.
    It would be nice to use Picard here, but there's a bug that makes it unusable. See:
    https://github.com/broadinstitute/picard/issues/284
    https://github.com/broadinstitute/picard/issues/160
    """
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/read_mapping/{filename}.no-pg-header-line.sam",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/read_mapping/{filename}.no-pg-header-line.sam"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/sams/{filename}.sam.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_variation_overviews:
    """Compare PAV and VCF nodes in variation_overview.txt."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/variation/variation_overview.no-dates.txt",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/variation/variation_overview.no-dates.txt"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/variation/variation_overview.txt.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_gene_classification_overviews:
    """Compare grouped genes in gene_classification_overview.txt."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/gene_classification/" \
            "gene_classification_overview.txt",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/gene_classification/" \
            "gene_classification_overview.txt"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/gene_classification/" \
            "gene_classification_overview.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_classified_groups:
    """Compare grouped genes in classified_groups.csv."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/gene_classification/" \
            "classified_groups.no-identifiers.csv",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/gene_classification/" \
            "classified_groups.no-identifiers.csv"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/gene_classification/" \
            "classified_groups.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_kmer_classification_overviews:
    """Compare kmer data in kmer_classification_overview.txt."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/kmer_classification/" \
            "kmer_classification_overview.txt",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/kmer_classification/" \
            "kmer_classification_overview.txt"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/kmer_classification/" \
            "kmer_classification_overview.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_pangenome_structure_overview:
    """Compage pangenome gene and kmer structure files."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/pangenome_size/{type}/{filename}",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/pangenome_size/{type}/{filename}"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/structure_{type}/{filename}.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_phenotype_overviews:
    """Compare phenotype statistics in phenotype_overview.txt."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/phenotype_overview.cleaned.txt",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/phenotype_overview.cleaned.txt"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/phenotype/phenotype_overview.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_grouping_overview:
    """Compare grouping overviews."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/group/grouping_overview.cleaned.txt",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/group/grouping_overview.cleaned.txt"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/group/grouping_overview.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_homology_groups:
    """Compare grouping output with pantools_homology_groups.txt."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/pantools_homology_groups.cleaned.txt",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/pantools_homology_groups.cleaned.txt"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/group/pantools_homology_groups.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_functional_classification_overview:
    """Compare functional classification overviews."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/function/functional_classification/{filename}.cleaned.txt",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/function/functional_classification/{filename}.cleaned.txt"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/functional_classification/" \
            "{filename}.txt.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_function_overview:
    """Compare functional annotation overviews."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/function/{filename}.no-identifiers.csv",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/function/{filename}.no-identifiers.csv"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/function_overview/{filename}.csv.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_msa_output:
    """Compare aggregated multiple sequence alignment output."""
    input:
        "output/aggregated/{dataset_name}_{dataset_type}/{version1}/msa/{subdir}/{filename}.aggregated.{ext}",
        "output/aggregated/{dataset_name}_{dataset_type}/{version2}/msa/{subdir}/{filename}.aggregated.{ext}"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/msa_{subdir}/{filename}.{ext}.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_go_enrichment_csv:
    """Compare output files for GO enrichment."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/function/go_enrichment/go_enrichment.sorted.csv",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/function/go_enrichment/go_enrichment.sorted.csv"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/go_enrichment/go_enrichment.csv.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_go_enrichment_overview:
    """Compare output files for GO enrichment."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/function/go_enrichment/go_enrichment_overview_per_go.cleaned.txt",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/function/go_enrichment/go_enrichment_overview_per_go.cleaned.txt"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/go_enrichment/go_enrichment_overview_per_go.txt.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_ani_distance_matrices:
    """Compare ANI distance matrices."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/ANI/MASH/ANI_distance_matrix.csv",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/ANI/MASH/ANI_distance_matrix.csv"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/phylogeny/ANI_distance_matrix.csv.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """

rule compare_core_phylogeny_sites_per_group:
    """Compare sites_per_group.csv created by core_phylogeny."""
    input:
        "output/databases/{dataset_name}_{dataset_type}/{version1}/core_phylogeny/sites_per_group.cleaned.csv",
        "output/databases/{dataset_name}_{dataset_type}/{version2}/core_phylogeny/sites_per_group.cleaned.csv"
    output:
        "output/validation/{dataset_name}_{dataset_type}/{version1}-vs-{version2}/phylogeny/sites_per_group.csv.cmp"
    priority: 100
    shell:
        """
        mkdir -p $(dirname {output})
        cmp {input} | tee {output}
        """
